﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;

namespace Persistence {

  [AttributeUsage(
    AttributeTargets.Field | AttributeTargets.Property,
    AllowMultiple=false,
    Inherited=false)]
  public class AutoStorableAttribute : Attribute {

    private string name = null;
    private object defaultValue = null;

    public string Name {
      get { return this.name; }
      set { this.name = value; }
    }

    public object DefaultValue {
      get { return this.defaultValue; }
      set { this.defaultValue = value; }
    }  

    private const BindingFlags instanceMembers =
      BindingFlags.Instance |
      BindingFlags.Public |
      BindingFlags.NonPublic;

    public static Dictionary<string, DataMemberAccessor> GetAutostorableAccessors(object obj) {
      Dictionary<string, DataMemberAccessor> autoStorableAccessors = 
        new Dictionary<string, DataMemberAccessor>();
      foreach (MemberInfo memberInfo in obj.GetType().GetMembers(instanceMembers)) {
        foreach (object attribute in memberInfo.GetCustomAttributes(false)) {          
          AutoStorableAttribute autoStorableAttribute =
            attribute as AutoStorableAttribute;
          if (autoStorableAttribute != null) {
            autoStorableAccessors.Add(memberInfo.Name,
              new DataMemberAccessor(memberInfo, autoStorableAttribute, obj));
          }
        }
      }
      return autoStorableAccessors;
    }    
  }

  public interface IDataMemberAccessor {
    string Name { get; }
    Type Type { get; }
    object DefaultValue { get; }
    object Get();
    void Set(object value);    
  }
    
  public class DataMemberAccessor : IDataMemberAccessor {

    public delegate object Getter();
    public delegate void Setter(object value);

    private string name;
    private Type type;
    private object defaultValue;
    private Getter getter;
    private Setter setter;

    public DataMemberAccessor(
        MemberInfo memberInfo,
        AutoStorableAttribute autoStorableAttribute,
        object obj) {
      if (memberInfo.MemberType == MemberTypes.Field) {
        this.getter = () => ((FieldInfo)memberInfo).GetValue(obj);
        this.setter = (value) => ((FieldInfo)memberInfo).SetValue(obj, value);
      } else if (memberInfo.MemberType == MemberTypes.Property) {
        this.getter = () => ((PropertyInfo)memberInfo).GetValue(obj, null);
        this.setter = (value) => ((PropertyInfo)memberInfo).SetValue(obj, value, null);
      } else {
        throw new NotSupportedException(
                "The AutoStorable attribute can only be applied to fields and properties.\n" +
                "However, it was applied to a member of type " + memberInfo.MemberType.ToString());
      }      
      this.name = autoStorableAttribute.Name != null ? autoStorableAttribute.Name : memberInfo.Name;
      this.type = memberInfo.ReflectedType;
      this.defaultValue = autoStorableAttribute.DefaultValue;                  
    }

    public DataMemberAccessor(
        string name, Type type, object defaultValue,
        Getter getter, Setter setter) {
      this.name = name;
      this.type = type;
      this.defaultValue = defaultValue;
      this.getter = getter;
      this.setter = setter;
    }

    public string Name {
      get { return this.name; }
    }

    public Type Type {
      get { return this.type; }
    }

    public object DefaultValue {
      get { return this.defaultValue; }
    }

    public object Get() {
      return this.getter();
    }

    public void Set(object value) {
      this.setter(value);
    }

    public override string ToString() {
      return String.Format("Property({0}, {1}, {2}, {3}, {4}",
        this.name,
        this.type == null ? "<null>" : this.type.FullName,
        this.defaultValue == null ? "<null>" : this.defaultValue,
        this.getter.Method,
        this.setter.Method);
    }
  }


  

}
